// ==================================================================== 
// 
// Fl_Combo.C
// 
//  Copyright (C) 1999 David Freese                                  
//                                                                    
//  This program is free software; you can redistribute it and/or     
//  modify itunder the terms of the GNU General Public License as    
//  published by the Free Software Foundation; either version 2 of    
//  the License, or (at your option) any later version.              
//                                                                    
//  This program is distributed in the hope that it will be useful,   
//  but WITHOUT ANY WARRANTY; without even the implied warranty of    
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     
//  GNU General Public License for more details.                     
//                                                                    
//  You should have received a copy of the GNU General Public License 
//  along with this program; if not, write to the Free Software       
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
//                                                                    
//  Author: Dave Freese ; debugged by Bill Spitzak
//  Internet: dfreese@intrepid.net                                   
//                                                                    
//====================================================================

#include "combo.H"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

// pointer to last widget to process an event
// needed to work around grab() disabling the passing of events
// to children of the grab() widget
// ONLY a single combobox is allowed to be active (in the dropdown state)
// at a time, so a static variable is sufficient.

static Fl_Widget* oldpushed;
static int wBtn = 17;

int winBrowser::handle (int e)
{
  int x0 = Fl::event_x_root(),
      y0 = Fl::event_y_root(),
      xmin = x(), xmax = x()+w(),
      ymin = y(), ymax = y()+h();
// is this event inside or outside of the combo box?      
  bool inside = (x0 > xmin && x0 < xmax) && (y0 > ymin && y0 < ymax);
  
// Status flag used to exit the infinite loop used to process events
// see code Fl_Combo::btn_cb_i () below
  if (e == FL_PUSH && !inside ||
      e == FL_KEYBOARD && Fl::event_key() == FL_Escape) {
    Status = CBO_CLOSED;
    return 1;
  }

// the dropdown box is a window which contains a single Fl_Hold_Browser
// widget.  the following code passes the current event to the correct
// child: Fl_Browser_ Fl_Box or Fl_Scrollbar
// and remembers which received the current event

  if ((e == FL_DRAG || e == FL_RELEASE) &&
      contains(oldpushed) && oldpushed != this) {
    oldpushed->handle(e);
  } else
    Bwsr->handle(e);
  oldpushed = Fl::pushed();
  return 1;
}

// process the result of the selection event in the dropdown browser

void winBrowser::browser_cb(Fl_Widget* w, void*) {
  winBrowser* W = (winBrowser*)(w->parent());
  W->Status = CBO_SELECTED;
}

// get or set the status of the dropdown box
// initial state is set to CBO_OPEN
// result of an event can set the state to either
// CBO_SELECTED - user selected an item from the browser
// CBO_CLOSED   - user clicked outside the browser box or
//              - user pressed Escape

int winBrowser::status()
{
  return (Status);
}


void winBrowser::status(int v)
{
  Status = v;
}

// Browser list box is populated on each call to show as the user
// or the program may have added, modified or deleted items in the
// list.  The list may also have been altered by a sort routine.

void winBrowser::show()
{
  Bwsr->clear ();
  for (int i = 0; i < sizelist; i++ )
    Bwsr->add(plist[i]);
  Fl_Window::show();
}

// class winBrowser creates the dropdown component of the combobox.

winBrowser::winBrowser (Fl_Widget *wc, int x, int y, int w, int h, const char *l)
    : Fl_Window(x,y,w,h,l)
{
//  plist = ((Fl_Combo *)(wc))->getlist();
//  sizelist = ((Fl_Combo *)(wc))->getsize();
// remove system window border
  clear_border();
  Bwsr = new cbo_browser (0, 0, w, h);
  Bwsr->Fl_Browser_::scrollbar_width (wBtn);
  Bwsr->callback (browser_cb);
}

// internal function to process the down arrow button callback
// create or show the dropdown browser window,
// set up the grab() ... process events ... release() loop
// update the Fl_Input widget component of combobox if user
// selected an item from the browser list

void Fl_Combo::btn_cb_i ()
{
  if (size < 1) return;  // empty list
  int vlines = (size < numlines) ? size : numlines;

// first time callback executed a new winBrowser is created
  if (wBwsr == NULL) {  
    wBwsr = new winBrowser (
      this,
      basew->x() + xpos, 
      basew->y() + ypos + height, 
      width + 2, 
      height * (vlines) + 2);
    wBwsr->end ();
  }

  wBwsr->plist = getlist ();
  wBwsr->sizelist = getsize ();
  
// winBrowser is now resized based on # lines & size of text font    

  wBwsr->resize(
    basew->x() + xpos, 
    basew->y() + ypos + height, 
    (size <= numlines) ? width - wBtn + 2 : width + 2, 
    (wBwsr->Bwsr->Fl_Browser::textsize() + 2) * (vlines) + 2);

  wBwsr->Bwsr->resize(
    0, 
    0, 
    (size <= numlines) ? width - wBtn + 2 : width + 2, 
    (wBwsr->Bwsr->Fl_Browser::textsize() + 2) * (vlines) + 2);

// Add vertical scroll bar only if # lines > design size

  (size <= numlines) ? 
    wBwsr->Bwsr->Fl_Browser_::has_scrollbar (0) :
    wBwsr->Bwsr->Fl_Browser_::has_scrollbar (Fl_Browser_::VERTICAL_ALWAYS);


  wBwsr->status(CBO_OPEN);

  Fl::grab (wBwsr);
  
  wBwsr->show();

  while (wBwsr->status() == CBO_OPEN)  Fl::wait ();

  wBwsr->hide ();
    
  Fl::release ();

  if (wBwsr->status() == CBO_SELECTED) {
    selected = wBwsr->Bwsr->value ();
    Input.value (wBwsr->Bwsr->text ( selected ) );
  }
  
  return;
}

// button callback has to execute the real callback contained in
// the parent Group widget

static void btn_cb (Fl_Widget *v, void *d)
{
  ((Fl_Combo *)(v->parent()))->btn_cb_i ();
  return;
}

Fl_Combo::Fl_Combo (int X,int Y,int W,int H, char *L)
 : Fl_Group (X, Y, W, H),
   Input (X, Y, W - wBtn, H, L),
   Btn (X + W - wBtn, Y, wBtn, H, "@#2>")
{
  xpos = X; 
  ypos = Y; 
  width = W; 
  height = H;
  size = 0;
  numlines = 6;
  maxsize = INITCBOSIZE;
  cbolist = new char*[INITCBOSIZE+1];

// basew pointer is used to allow the dropdown browser window to be
// properly placed independent of application resize or movement
  basew = this;
  while (basew->parent() != NULL)
    basew = basew->parent();
  
  Btn.labeltype (FL_SYMBOL_LABEL);
  Btn.callback ((Fl_Callback *)btn_cb, 0);
    
  clear_cbo();

  end();
}

// default dropdown button is triangle pointing down
// user may specify any one of the "symbol label" drawings

void Fl_Combo::ButtonLabel( const char *lbl )
{
  Btn.label (lbl);
  Btn.labeltype (FL_SYMBOL_LABEL);
}
 
// Combobox value is contained in the Input widget

void Fl_Combo::value( const char *s )
{
  Input.value (s);
}

const char *Fl_Combo::value()
{
  return (Input.value ());
}

void Fl_Combo::align(int a)
{
  Input.align (a);
}

void Fl_Combo::labelcolor(int c)
{
  Input.labelcolor (c);
}

void Fl_Combo::textfont(int tf)
{
  Input.textfont (tf);
}
 
void Fl_Combo::textsize(int sz)
{
  Input.textsize (sz);
}

// the list values are contained in the dynamically allocated 
// char *cbolist[]

static char **temp;

void Fl_Combo::incr_array() {
  if ( size < maxsize ) return;         // less than max size
  maxsize += INITCBOSIZE;               // increment array by INITCBOSIZE 
  temp = new char*[maxsize + 1];
  for (int i = 0; i <= size; i++) temp[i] = cbolist[i];
  delete[] cbolist;
  cbolist = temp;
}


void Fl_Combo::clear_cbo()
{
  if( size > 0 )
    for( int i = 1; i < size ; i++ ) {
      if( cbolist[i] ) delete[] cbolist[i];
      cbolist[i] = 0;
    }
  size = 0;
  selected = 0;
}

static int  Compare( const void *s1, const void *s2 )
{
  char **str1 = (char **) (s1),
       **str2 = (char **) (s2);
  return( strcmp( *str1, *str2 ) );
}

void Fl_Combo::sort()
{
  if( size > 1 )
    qsort( &cbolist[1], size - 1, sizeof( char * ), Compare );
}

void Fl_Combo::add( const char *s )
{
// no empty strings allowed in pulldown list
  if (strlen (s) == 0)
    return;
// no duplicate entries (no case sensitivity) allowed in pulldown list
  for (int i = 0; i < size; i++) {
    if (strcasecmp (cbolist[i],s) == 0)
      return;
  }
  incr_array();
  cbolist[size] = new char[ strlen(s) + 1];
  strcpy( cbolist[size], s );
  size++;
}

// end of Fl_Combo.C
