// ====================================================================
//
//  date.cc  date class for Fast Light Took Kit Checkbook Program
//
//  Copyright (C) 1998 David Freese                                  
//                                                                   
//  This program is free software; you can redistribute it and/or    
//  modify it under the terms of the GNU General Public License as   
//  published by the Free Software Foundation; either version 2 of   
//  the License, or (at your option) any later version.              
//                                                                   
//  This program is distributed in the hope that it will be useful,  
//  but WITHOUT ANY WARRANTY; without even the implied warranty of   
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
//  GNU General Public License for more details.                     
//                                                                   
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software      
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
//                                                                   
//    Author: Dave Freese                                            
//  Internet: dfreese@intrepid.net                                   
//                                                                   
// ====================================================================

#include "date.H"
#include <iostream.h>
#include <iomanip.h>

const int Date::days[] = 
  { 0, 31, 28, 31, 30,  31,  30,  31,  31,  30,  31,  30, 31 };
const int Date::jdays[] = 
  { 0,  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };

void Date::today()
{
  time_t t;
  struct tm *now;
  time( &t );
  now = localtime( &t );
  year = now->tm_year + 1900;
  month = now->tm_mon + 1;
  day = now->tm_mday;
}

Date::Date()
{
  today();
  fmt = 0;
}

Date::Date( int m, int d, int y )
{
  setDate( m, d, y );
  fmt = 0;
}

void Date::setDate( int mm, int dd, int yy )
{
  if( isvalid( mm, dd, yy ) ) {
    year = yy; month = mm; day = dd;
  } else
    today();
}

void Date::setDate( Date &dt )
{
  year = dt.year;
  month = dt.month;
  day = dt.day;
}

void Date::setFormat( int iFmt )
{
  fmt = iFmt;
}

void Date::Year( int y )
{
  year = y;
}

int Date::Year()
{
  return year;
}

void Date::Month( int m )
{
  month = m;
}

int Date::Month()
{
  return month;
}

void Date::Day( int d )
{
  day = d;
}

int Date::Day()
{
  return day;
}

bool Date::leapYear( int y )
{
  if( y % 400 == 0 || ( y % 100 != 0 && y % 4 == 0 ) )
    return true;
  return false;
}


bool Date::isvalid( int m, int d, int y )
{
  if( y < 1970 || y > 2035 ) return false;
  if( m < 1 || m > 12 ) return false;
  if( d < 1 ) return false;
  if( leapYear( y ) ){
    if( m == 2 && d > 29 )
      return false;
    else
      return true;
  }
  if( d > days[m] ) return false;
  return true;
}

ostream &operator<<( ostream &output, const Date &d )
{
  static char *monthName[13] = { "", 
         "Jan", "Feb", "Mar", "Apr", "May", "Jun",
         "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
  switch (d.fmt) {
    case 1 :
      output << monthName[ d.month ] << ' ' << d.day << ", " << d.year;
      break;
    case 0 :
    default :
      output << d.year << '/' ;
      cout.setf( ios::right, ios::adjustfield);
      cout.fill('0');
      output << setw(2) << d.month << '/' << d.day;
      break;
  }
  return output;
}

char *Date::szDate (int n = 2)
{
  static char s[11];
  Date today;
  switch (n) {
    case 4 :
      sprintf (s, "%02d/%02d/%04d", Month(), Day(), Year() );
      break;   
    case 2 :
    default :
      sprintf( s,"%02d/%02d/%02d", Month(), Day(),
               Year() < 2000 ? Year() - 1900 : Year() - 2000 );
      break;
  }      
  return s;
}

bool Date::endOfMonth( int d )
{
  if( month == 2 && leapYear( year ) )
    return (d == 29 );  // last day of Feb in leap year
  else
    return (d == days[ month ] );
}

void Date::helpIncrement()
{
  if( endOfMonth( day ) && month == 12 ) {  // end year
    day = 1;
    month = 1;
    ++year;
  } else if ( endOfMonth( day ) ) {
    day = 1;
    ++month;
  } else
    ++day;
}

Date &Date::operator++()
{
  helpIncrement();
  return *this;     // reference return to create an lvalue
}

Date Date::operator++( int )
{
  Date temp = *this;
  helpIncrement();
  return temp;     // return non-increment, saved temporary object
}

const Date &Date::operator+=( int ndays )
{
  for( int i = 0; i < ndays; i++ )
    helpIncrement();
  return *this;    // enables cascading
}


bool Date::operator==( const Date &d )
{
  if( this->year != d.year ) return false;
  if( this->month != d.month ) return false;
  if( this->day != d.day ) return false;
  return true;
}

bool Date::operator!=( const Date &d )
{
  return ( !( *this == d ) );
}

bool Date::operator<( const Date &d )
{
  if( this->year < d.year ) return true;
  if( this->year > d.year ) return false;
  if( this->month < d.month ) return true;
  if( this->month > d.month ) return false;
  if( this->day < d.day ) return true;
  return false;
}

bool Date::operator>( const Date &d )
{
  if( *this < d ) return false;
  if( *this == d ) return false;
  return true;
}

void Date::operator=( const Date &d )
{
  this->year = d.year;
  this->month = d.month;
  this->day = d.day;
}

double Date::dJulian()
{
  int DaysInYear = 365,
      JulianDay = 0;
  if( leapYear( year ) ) DaysInYear++;
  JulianDay = jdays[month] + day;
  if( leapYear( year ) &&  month > 2 )
    JulianDay++;
  return( year + 1.0 * (JulianDay-1) / DaysInYear );
}

